@@ -1,17 +1,12 @@
-from typing import List, Optional
+import logging
+import typing
 
+import hpack
+import hyperframe.frame
 import pytest
 import trio as concurrency
 
-from httpcore import (
-    AsyncConnectionPool,
-    ConnectError,
-    PoolTimeout,
-    ReadError,
-    UnsupportedProtocol,
-)
-from httpcore.backends.base import AsyncNetworkStream
-from httpcore.backends.mock import AsyncMockBackend
+import httpcore
 
 
 @pytest.mark.anyio
@@ -19,7 +14,7 @@ async def test_connection_pool_with_keepalive():
     """
     By default HTTP/1.1 requests should be returned to the connection pool.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -34,7 +29,7 @@ async def test_connection_pool_with_keepalive():
         ]
     )
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         network_backend=network_backend,
     ) as pool:
         # Sending an intial request, which once complete will return to the pool, IDLE.
@@ -43,6 +38,10 @@ async def test_connection_pool_with_keepalive():
             assert info == [
                 "<AsyncHTTPConnection ['https://example.com:443', HTTP/1.1, ACTIVE, Request Count: 1]>"
             ]
+            assert (
+                repr(pool)
+                == "<AsyncConnectionPool [Requests: 1 active, 0 queued | Connections: 1 active, 0 idle]>"
+            )
             await response.aread()
 
         assert response.status == 200
@@ -51,6 +50,10 @@ async def test_connection_pool_with_keepalive():
         assert info == [
             "<AsyncHTTPConnection ['https://example.com:443', HTTP/1.1, IDLE, Request Count: 1]>"
         ]
+        assert (
+            repr(pool)
+            == "<AsyncConnectionPool [Requests: 0 active, 0 queued | Connections: 0 active, 1 idle]>"
+        )
 
         # Sending a second request to the same origin will reuse the existing IDLE connection.
         async with pool.stream("GET", "https://example.com/") as response:
@@ -58,6 +61,10 @@ async def test_connection_pool_with_keepalive():
             assert info == [
                 "<AsyncHTTPConnection ['https://example.com:443', HTTP/1.1, ACTIVE, Request Count: 2]>"
             ]
+            assert (
+                repr(pool)
+                == "<AsyncConnectionPool [Requests: 1 active, 0 queued | Connections: 1 active, 0 idle]>"
+            )
             await response.aread()
 
         assert response.status == 200
@@ -66,23 +73,35 @@ async def test_connection_pool_with_keepalive():
         assert info == [
             "<AsyncHTTPConnection ['https://example.com:443', HTTP/1.1, IDLE, Request Count: 2]>"
         ]
+        assert (
+            repr(pool)
+            == "<AsyncConnectionPool [Requests: 0 active, 0 queued | Connections: 0 active, 1 idle]>"
+        )
 
         # Sending a request to a different origin will not reuse the existing IDLE connection.
         async with pool.stream("GET", "http://example.com/") as response:
             info = [repr(c) for c in pool.connections]
             assert info == [
-                "<AsyncHTTPConnection ['http://example.com:80', HTTP/1.1, ACTIVE, Request Count: 1]>",
                 "<AsyncHTTPConnection ['https://example.com:443', HTTP/1.1, IDLE, Request Count: 2]>",
+                "<AsyncHTTPConnection ['http://example.com:80', HTTP/1.1, ACTIVE, Request Count: 1]>",
             ]
+            assert (
+                repr(pool)
+                == "<AsyncConnectionPool [Requests: 1 active, 0 queued | Connections: 1 active, 1 idle]>"
+            )
             await response.aread()
 
         assert response.status == 200
         assert response.content == b"Hello, world!"
         info = [repr(c) for c in pool.connections]
         assert info == [
-            "<AsyncHTTPConnection ['http://example.com:80', HTTP/1.1, IDLE, Request Count: 1]>",
             "<AsyncHTTPConnection ['https://example.com:443', HTTP/1.1, IDLE, Request Count: 2]>",
+            "<AsyncHTTPConnection ['http://example.com:80', HTTP/1.1, IDLE, Request Count: 1]>",
         ]
+        assert (
+            repr(pool)
+            == "<AsyncConnectionPool [Requests: 0 active, 0 queued | Connections: 0 active, 2 idle]>"
+        )
 
 
 @pytest.mark.anyio
@@ -91,7 +110,7 @@ async def test_connection_pool_with_close():
     HTTP/1.1 requests that include a 'Connection: Close' header should
     not be returned to the connection pool.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -101,7 +120,7 @@ async def test_connection_pool_with_close():
         ]
     )
 
-    async with AsyncConnectionPool(network_backend=network_backend) as pool:
+    async with httpcore.AsyncConnectionPool(network_backend=network_backend) as pool:
         # Sending an intial request, which once complete will not return to the pool.
         async with pool.stream(
             "GET", "https://example.com/", headers={"Connection": "close"}
@@ -118,13 +137,130 @@ async def test_connection_pool_with_close():
         assert info == []
 
 
+@pytest.mark.anyio
+async def test_connection_pool_with_http2():
+    """
+    Test a connection pool with HTTP/2 requests.
+    """
+    network_backend = httpcore.AsyncMockBackend(
+        buffer=[
+            hyperframe.frame.SettingsFrame().serialize(),
+            hyperframe.frame.HeadersFrame(
+                stream_id=1,
+                data=hpack.Encoder().encode(
+                    [
+                        (b":status", b"200"),
+                        (b"content-type", b"plain/text"),
+                    ]
+                ),
+                flags=["END_HEADERS"],
+            ).serialize(),
+            hyperframe.frame.DataFrame(
+                stream_id=1, data=b"Hello, world!", flags=["END_STREAM"]
+            ).serialize(),
+            hyperframe.frame.HeadersFrame(
+                stream_id=3,
+                data=hpack.Encoder().encode(
+                    [
+                        (b":status", b"200"),
+                        (b"content-type", b"plain/text"),
+                    ]
+                ),
+                flags=["END_HEADERS"],
+            ).serialize(),
+            hyperframe.frame.DataFrame(
+                stream_id=3, data=b"Hello, world!", flags=["END_STREAM"]
+            ).serialize(),
+        ],
+        http2=True,
+    )
+
+    async with httpcore.AsyncConnectionPool(
+        network_backend=network_backend,
+    ) as pool:
+        # Sending an intial request, which once complete will return to the pool, IDLE.
+        response = await pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<AsyncHTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 1]>"
+        ]
+
+        # Sending a second request to the same origin will reuse the existing IDLE connection.
+        response = await pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<AsyncHTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 2]>"
+        ]
+
+
+@pytest.mark.anyio
+async def test_connection_pool_with_http2_goaway():
+    """
+    Test a connection pool with HTTP/2 requests, that cleanly disconnects
+    with a GoAway frame after the first request.
+    """
+    network_backend = httpcore.AsyncMockBackend(
+        buffer=[
+            hyperframe.frame.SettingsFrame().serialize(),
+            hyperframe.frame.HeadersFrame(
+                stream_id=1,
+                data=hpack.Encoder().encode(
+                    [
+                        (b":status", b"200"),
+                        (b"content-type", b"plain/text"),
+                    ]
+                ),
+                flags=["END_HEADERS"],
+            ).serialize(),
+            hyperframe.frame.DataFrame(
+                stream_id=1, data=b"Hello, world!", flags=["END_STREAM"]
+            ).serialize(),
+            hyperframe.frame.GoAwayFrame(
+                stream_id=0, error_code=0, last_stream_id=1
+            ).serialize(),
+            b"",
+        ],
+        http2=True,
+    )
+
+    async with httpcore.AsyncConnectionPool(
+        network_backend=network_backend,
+    ) as pool:
+        # Sending an intial request, which once complete will return to the pool, IDLE.
+        response = await pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<AsyncHTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 1]>"
+        ]
+
+        # Sending a second request to the same origin will require a new connection.
+        # The original connection has now been closed.
+        response = await pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<AsyncHTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 1]>",
+        ]
+
+
 @pytest.mark.anyio
 async def test_trace_request():
     """
     The 'trace' request extension allows for a callback function to inspect the
     internal events that occur while sending a request.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -139,7 +275,7 @@ async def test_trace_request():
     async def trace(name, kwargs):
         called.append(name)
 
-    async with AsyncConnectionPool(network_backend=network_backend) as pool:
+    async with httpcore.AsyncConnectionPool(network_backend=network_backend) as pool:
         await pool.request("GET", "https://example.com/", extensions={"trace": trace})
 
     assert called == [
@@ -160,20 +296,88 @@ async def test_trace_request():
     ]
 
 
+@pytest.mark.anyio
+async def test_debug_request(caplog):
+    """
+    The 'trace' request extension allows for a callback function to inspect the
+    internal events that occur while sending a request.
+    """
+    caplog.set_level(logging.DEBUG)
+
+    network_backend = httpcore.AsyncMockBackend(
+        [
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+        ]
+    )
+
+    async with httpcore.AsyncConnectionPool(network_backend=network_backend) as pool:
+        await pool.request("GET", "http://example.com/")
+
+    assert caplog.record_tuples == [
+        (
+            "httpcore.connection",
+            logging.DEBUG,
+            "connect_tcp.started host='example.com' port=80 local_address=None timeout=None socket_options=None",
+        ),
+        (
+            "httpcore.connection",
+            logging.DEBUG,
+            "connect_tcp.complete return_value=<httpcore.AsyncMockStream>",
+        ),
+        (
+            "httpcore.http11",
+            logging.DEBUG,
+            "send_request_headers.started request=<Request [b'GET']>",
+        ),
+        ("httpcore.http11", logging.DEBUG, "send_request_headers.complete"),
+        (
+            "httpcore.http11",
+            logging.DEBUG,
+            "send_request_body.started request=<Request [b'GET']>",
+        ),
+        ("httpcore.http11", logging.DEBUG, "send_request_body.complete"),
+        (
+            "httpcore.http11",
+            logging.DEBUG,
+            "receive_response_headers.started request=<Request [b'GET']>",
+        ),
+        (
+            "httpcore.http11",
+            logging.DEBUG,
+            "receive_response_headers.complete return_value="
+            "(b'HTTP/1.1', 200, b'OK', [(b'Content-Type', b'plain/text'), (b'Content-Length', b'13')])",
+        ),
+        (
+            "httpcore.http11",
+            logging.DEBUG,
+            "receive_response_body.started request=<Request [b'GET']>",
+        ),
+        ("httpcore.http11", logging.DEBUG, "receive_response_body.complete"),
+        ("httpcore.http11", logging.DEBUG, "response_closed.started"),
+        ("httpcore.http11", logging.DEBUG, "response_closed.complete"),
+        ("httpcore.connection", logging.DEBUG, "close.started"),
+        ("httpcore.connection", logging.DEBUG, "close.complete"),
+    ]
+
+
 @pytest.mark.anyio
 async def test_connection_pool_with_http_exception():
     """
     HTTP/1.1 requests that result in an exception during the connection should
     not be returned to the connection pool.
     """
-    network_backend = AsyncMockBackend([b"Wait, this isn't valid HTTP!"])
+    network_backend = httpcore.AsyncMockBackend([b"Wait, this isn't valid HTTP!"])
 
     called = []
 
     async def trace(name, kwargs):
         called.append(name)
 
-    async with AsyncConnectionPool(network_backend=network_backend) as pool:
+    async with httpcore.AsyncConnectionPool(network_backend=network_backend) as pool:
         # Sending an initial request, which once complete will not return to the pool.
         with pytest.raises(Exception):
             await pool.request(
@@ -206,15 +410,18 @@ async def test_connection_pool_with_connect_exception():
     be returned to the connection pool.
     """
 
-    class FailedConnectBackend(AsyncMockBackend):
+    class FailedConnectBackend(httpcore.AsyncMockBackend):
         async def connect_tcp(
             self,
             host: str,
             port: int,
-            timeout: Optional[float] = None,
-            local_address: Optional[str] = None,
-        ) -> AsyncNetworkStream:
-            raise ConnectError("Could not connect")
+            timeout: typing.Optional[float] = None,
+            local_address: typing.Optional[str] = None,
+            socket_options: typing.Optional[
+                typing.Iterable[httpcore.SOCKET_OPTION]
+            ] = None,
+        ) -> httpcore.AsyncNetworkStream:
+            raise httpcore.ConnectError("Could not connect")
 
     network_backend = FailedConnectBackend([])
 
@@ -223,7 +430,7 @@ async def test_connection_pool_with_connect_exception():
     async def trace(name, kwargs):
         called.append(name)
 
-    async with AsyncConnectionPool(network_backend=network_backend) as pool:
+    async with httpcore.AsyncConnectionPool(network_backend=network_backend) as pool:
         # Sending an initial request, which once complete will not return to the pool.
         with pytest.raises(Exception):
             await pool.request(
@@ -245,7 +452,7 @@ async def test_connection_pool_with_immediate_expiry():
     Connection pools with keepalive_expiry=0.0 should immediately expire
     keep alive connections.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -255,7 +462,7 @@ async def test_connection_pool_with_immediate_expiry():
         ]
     )
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         keepalive_expiry=0.0,
         network_backend=network_backend,
     ) as pool:
@@ -279,7 +486,7 @@ async def test_connection_pool_with_no_keepalive_connections_allowed():
     When 'max_keepalive_connections=0' is used, IDLE connections should not
     be returned to the pool.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -289,7 +496,7 @@ async def test_connection_pool_with_no_keepalive_connections_allowed():
         ]
     )
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         max_keepalive_connections=0, network_backend=network_backend
     ) as pool:
         # Sending an intial request, which once complete will not return to the pool.
@@ -312,7 +519,7 @@ async def test_connection_pool_concurrency():
     HTTP/1.1 requests made in concurrency must not ever exceed the maximum number
     of allowable connection in the pool.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -328,10 +535,10 @@ async def test_connection_pool_concurrency():
             info_list.append(info)
             await response.aread()
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         max_connections=1, network_backend=network_backend
     ) as pool:
-        info_list: List[str] = []
+        info_list: typing.List[str] = []
         async with concurrency.open_nursery() as nursery:
             for domain in ["a.com", "b.com", "c.com", "d.com", "e.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
@@ -357,7 +564,7 @@ async def test_connection_pool_concurrency_same_domain_closing():
     HTTP/1.1 requests made in concurrency must not ever exceed the maximum number
     of allowable connection in the pool.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -374,10 +581,10 @@ async def test_connection_pool_concurrency_same_domain_closing():
             info_list.append(info)
             await response.aread()
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         max_connections=1, network_backend=network_backend, http2=True
     ) as pool:
-        info_list: List[str] = []
+        info_list: typing.List[str] = []
         async with concurrency.open_nursery() as nursery:
             for domain in ["a.com", "a.com", "a.com", "a.com", "a.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
@@ -399,7 +606,7 @@ async def test_connection_pool_concurrency_same_domain_keepalive():
     HTTP/1.1 requests made in concurrency must not ever exceed the maximum number
     of allowable connection in the pool.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -416,10 +623,10 @@ async def test_connection_pool_concurrency_same_domain_keepalive():
             info_list.append(info)
             await response.aread()
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         max_connections=1, network_backend=network_backend, http2=True
     ) as pool:
-        info_list: List[str] = []
+        info_list: typing.List[str] = []
         async with concurrency.open_nursery() as nursery:
             for domain in ["a.com", "a.com", "a.com", "a.com", "a.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
@@ -437,14 +644,19 @@ async def test_connection_pool_concurrency_same_domain_keepalive():
                 "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 5]>",
             ]
 
+    assert (
+        repr(pool)
+        == "<AsyncConnectionPool [Requests: 0 active, 0 queued | Connections: 0 active, 0 idle]>"
+    )
+
 
 @pytest.mark.anyio
 async def test_unsupported_protocol():
-    async with AsyncConnectionPool() as pool:
-        with pytest.raises(UnsupportedProtocol):
+    async with httpcore.AsyncConnectionPool() as pool:
+        with pytest.raises(httpcore.UnsupportedProtocol):
             await pool.request("GET", "ftp://www.example.com/")
 
-        with pytest.raises(UnsupportedProtocol):
+        with pytest.raises(httpcore.UnsupportedProtocol):
             await pool.request("GET", "://www.example.com/")
 
 
@@ -454,7 +666,7 @@ async def test_connection_pool_closed_while_request_in_flight():
     Closing a connection pool while a request/response is still in-flight
     should raise an error.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -464,14 +676,14 @@ async def test_connection_pool_closed_while_request_in_flight():
         ]
     )
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         network_backend=network_backend,
     ) as pool:
         # Send a request, and then close the connection pool while the
         # response has not yet been streamed.
         async with pool.stream("GET", "https://example.com/") as response:
             await pool.aclose()
-            with pytest.raises(ReadError):
+            with pytest.raises(httpcore.ReadError):
                 await response.aread()
 
 
@@ -480,7 +692,7 @@ async def test_connection_pool_timeout():
     """
     Ensure that exceeding max_connections can cause a request to timeout.
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -490,18 +702,80 @@ async def test_connection_pool_timeout():
         ]
     )
 
-    async with AsyncConnectionPool(
+    async with httpcore.AsyncConnectionPool(
         network_backend=network_backend, max_connections=1
     ) as pool:
         # Send a request to a pool that is configured to only support a single
         # connection, and then ensure that a second concurrent request
         # fails with a timeout.
         async with pool.stream("GET", "https://example.com/"):
-            with pytest.raises(PoolTimeout):
+            with pytest.raises(httpcore.PoolTimeout):
                 extensions = {"timeout": {"pool": 0.0001}}
                 await pool.request("GET", "https://example.com/", extensions=extensions)
 
 
+@pytest.mark.anyio
+async def test_connection_pool_timeout_zero():
+    """
+    A pool timeout of 0 shouldn't raise a PoolTimeout if there's
+    no need to wait on a new connection.
+    """
+    network_backend = httpcore.AsyncMockBackend(
+        [
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+        ]
+    )
+
+    # Use a pool timeout of zero.
+    extensions = {"timeout": {"pool": 0}}
+
+    # A connection pool configured to allow only one connection at a time.
+    async with httpcore.AsyncConnectionPool(
+        network_backend=network_backend, max_connections=1
+    ) as pool:
+        # Two consecutive requests with a pool timeout of zero.
+        # Both succeed without raising a timeout.
+        response = await pool.request(
+            "GET", "https://example.com/", extensions=extensions
+        )
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        response = await pool.request(
+            "GET", "https://example.com/", extensions=extensions
+        )
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+    # A connection pool configured to allow only one connection at a time.
+    async with httpcore.AsyncConnectionPool(
+        network_backend=network_backend, max_connections=1
+    ) as pool:
+        # Two concurrent requests with a pool timeout of zero.
+        # Only the first will succeed without raising a timeout.
+        async with pool.stream(
+            "GET", "https://example.com/", extensions=extensions
+        ) as response:
+            # The first response hasn't yet completed.
+            with pytest.raises(httpcore.PoolTimeout):
+                # So a pool timeout occurs.
+                await pool.request("GET", "https://example.com/", extensions=extensions)
+            # The first response now completes.
+            await response.aread()
+
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+
 @pytest.mark.anyio
 async def test_http11_upgrade_connection():
     """
@@ -513,7 +787,7 @@ async def test_http11_upgrade_connection():
     https://httpwg.org/specs/rfc9110.html#status.101
     https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/101
     """
-    network_backend = AsyncMockBackend(
+    network_backend = httpcore.AsyncMockBackend(
         [
             b"HTTP/1.1 101 Switching Protocols\r\n",
             b"Connection: upgrade\r\n",
@@ -522,15 +796,37 @@ async def test_http11_upgrade_connection():
             b"...",
         ]
     )
-    async with AsyncConnectionPool(
+
+    called = []
+
+    async def trace(name, kwargs):
+        called.append(name)
+
+    async with httpcore.AsyncConnectionPool(
         network_backend=network_backend, max_connections=1
     ) as pool:
         async with pool.stream(
             "GET",
             "wss://example.com/",
             headers={"Connection": "upgrade", "Upgrade": "custom"},
+            extensions={"trace": trace},
         ) as response:
             assert response.status == 101
             network_stream = response.extensions["network_stream"]
             content = await network_stream.read(max_bytes=1024)
             assert content == b"..."
+
+    assert called == [
+        "connection.connect_tcp.started",
+        "connection.connect_tcp.complete",
+        "connection.start_tls.started",
+        "connection.start_tls.complete",
+        "http11.send_request_headers.started",
+        "http11.send_request_headers.complete",
+        "http11.send_request_body.started",
+        "http11.send_request_body.complete",
+        "http11.receive_response_headers.started",
+        "http11.receive_response_headers.complete",
+        "http11.response_closed.started",
+        "http11.response_closed.complete",
+    ]
