@@ -1,19 +1,12 @@
 import pytest
 
-from httpcore import (
-    HTTP11Connection,
-    ConnectionNotAvailable,
-    LocalProtocolError,
-    Origin,
-    RemoteProtocolError,
-)
-from httpcore.backends.mock import MockStream
+import httpcore
 
 
 
 def test_http11_connection():
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -22,7 +15,7 @@ def test_http11_connection():
             b"Hello, world!",
         ]
     )
-    with HTTP11Connection(
+    with httpcore.HTTP11Connection(
         origin=origin, stream=stream, keepalive_expiry=5.0
     ) as conn:
         response = conn.request("GET", "https://example.com/")
@@ -45,8 +38,8 @@ def test_http11_connection_unread_response():
     If the client releases the response without reading it to termination,
     then the connection will not be reusable.
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -55,7 +48,7 @@ def test_http11_connection_unread_response():
             b"Hello, world!",
         ]
     )
-    with HTTP11Connection(origin=origin, stream=stream) as conn:
+    with httpcore.HTTP11Connection(origin=origin, stream=stream) as conn:
         with conn.stream("GET", "https://example.com/") as response:
             assert response.status == 200
 
@@ -75,10 +68,10 @@ def test_http11_connection_with_remote_protocol_error():
     If a remote protocol error occurs, then no response will be returned,
     and the connection will not be reusable.
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream([b"Wait, this isn't valid HTTP!", b""])
-    with HTTP11Connection(origin=origin, stream=stream) as conn:
-        with pytest.raises(RemoteProtocolError):
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream([b"Wait, this isn't valid HTTP!", b""])
+    with httpcore.HTTP11Connection(origin=origin, stream=stream) as conn:
+        with pytest.raises(httpcore.RemoteProtocolError):
             conn.request("GET", "https://example.com/")
 
         assert not conn.is_idle()
@@ -96,8 +89,8 @@ def test_http11_connection_with_incomplete_response():
     """
     We should be gracefully handling the case where the connection ends prematurely.
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -106,8 +99,8 @@ def test_http11_connection_with_incomplete_response():
             b"Hello, wor",
         ]
     )
-    with HTTP11Connection(origin=origin, stream=stream) as conn:
-        with pytest.raises(RemoteProtocolError):
+    with httpcore.HTTP11Connection(origin=origin, stream=stream) as conn:
+        with pytest.raises(httpcore.RemoteProtocolError):
             conn.request("GET", "https://example.com/")
 
         assert not conn.is_idle()
@@ -126,8 +119,8 @@ def test_http11_connection_with_local_protocol_error():
     If a local protocol error occurs, then no response will be returned,
     and the connection will not be reusable.
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -136,8 +129,8 @@ def test_http11_connection_with_local_protocol_error():
             b"Hello, world!",
         ]
     )
-    with HTTP11Connection(origin=origin, stream=stream) as conn:
-        with pytest.raises(LocalProtocolError) as exc_info:
+    with httpcore.HTTP11Connection(origin=origin, stream=stream) as conn:
+        with pytest.raises(httpcore.LocalProtocolError) as exc_info:
             conn.request("GET", "https://example.com/", headers={"Host": "\0"})
 
         assert str(exc_info.value) == "Illegal header value b'\\x00'"
@@ -158,8 +151,8 @@ def test_http11_connection_handles_one_active_request():
     Attempting to send a request while one is already in-flight will raise
     a ConnectionNotAvailable exception.
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -168,9 +161,9 @@ def test_http11_connection_handles_one_active_request():
             b"Hello, world!",
         ]
     )
-    with HTTP11Connection(origin=origin, stream=stream) as conn:
+    with httpcore.HTTP11Connection(origin=origin, stream=stream) as conn:
         with conn.stream("GET", "https://example.com/"):
-            with pytest.raises(ConnectionNotAvailable):
+            with pytest.raises(httpcore.ConnectionNotAvailable):
                 conn.request("GET", "https://example.com/")
 
 
@@ -179,8 +172,8 @@ def test_http11_connection_attempt_close():
     """
     A connection can only be closed when it is idle.
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -189,7 +182,7 @@ def test_http11_connection_attempt_close():
             b"Hello, world!",
         ]
     )
-    with HTTP11Connection(origin=origin, stream=stream) as conn:
+    with httpcore.HTTP11Connection(origin=origin, stream=stream) as conn:
         with conn.stream("GET", "https://example.com/") as response:
             response.read()
             assert response.status == 200
@@ -201,9 +194,9 @@ def test_http11_request_to_incorrect_origin():
     """
     A connection can only send requests to whichever origin it is connected to.
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream([])
-    with HTTP11Connection(origin=origin, stream=stream) as conn:
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream([])
+    with httpcore.HTTP11Connection(origin=origin, stream=stream) as conn:
         with pytest.raises(RuntimeError):
             conn.request("GET", "https://other.com/")
 
@@ -217,8 +210,8 @@ def test_http11_expect_continue():
     https://httpwg.org/specs/rfc9110.html#status.100
     https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/100
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 100 Continue\r\n",
             b"\r\n",
@@ -229,7 +222,7 @@ def test_http11_expect_continue():
             b"Hello, world!",
         ]
     )
-    with HTTP11Connection(
+    with httpcore.HTTP11Connection(
         origin=origin, stream=stream, keepalive_expiry=5.0
     ) as conn:
         response = conn.request(
@@ -252,8 +245,8 @@ def test_http11_upgrade_connection():
     https://httpwg.org/specs/rfc9110.html#status.101
     https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/101
     """
-    origin = Origin(b"wss", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"wss", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 101 Switching Protocols\r\n",
             b"Connection: upgrade\r\n",
@@ -262,7 +255,7 @@ def test_http11_upgrade_connection():
             b"...",
         ]
     )
-    with HTTP11Connection(
+    with httpcore.HTTP11Connection(
         origin=origin, stream=stream, keepalive_expiry=5.0
     ) as conn:
         with conn.stream(
@@ -277,6 +270,57 @@ def test_http11_upgrade_connection():
 
 
 
+def test_http11_upgrade_with_trailing_data():
+    """
+    HTTP "101 Switching Protocols" indicates an upgraded connection.
+
+    In `CONNECT` and `Upgrade:` requests, we need to handover the trailing data
+    in the h11.Connection object.
+
+    https://h11.readthedocs.io/en/latest/api.html#switching-protocols
+    """
+    origin = httpcore.Origin(b"wss", b"example.com", 443)
+    stream = httpcore.MockStream(
+        # The first element of this mock network stream buffer simulates networking
+        # in which response headers and data are received at once.
+        # This means that "foobar" becomes trailing data.
+        [
+            (
+                b"HTTP/1.1 101 Switching Protocols\r\n"
+                b"Connection: upgrade\r\n"
+                b"Upgrade: custom\r\n"
+                b"\r\n"
+                b"foobar"
+            ),
+            b"baz",
+        ]
+    )
+    with httpcore.HTTP11Connection(
+        origin=origin, stream=stream, keepalive_expiry=5.0
+    ) as conn:
+        with conn.stream(
+            "GET",
+            "wss://example.com/",
+            headers={"Connection": "upgrade", "Upgrade": "custom"},
+        ) as response:
+            assert response.status == 101
+            network_stream = response.extensions["network_stream"]
+
+            content = network_stream.read(max_bytes=3)
+            assert content == b"foo"
+            content = network_stream.read(max_bytes=3)
+            assert content == b"bar"
+            content = network_stream.read(max_bytes=3)
+            assert content == b"baz"
+
+            # Lazy tests for HTTP11UpgradeStream
+            network_stream.write(b"spam")
+            invalid = network_stream.get_extra_info("invalid")
+            assert invalid is None
+            network_stream.close()
+
+
+
 def test_http11_early_hints():
     """
     HTTP "103 Early Hints" is an interim response.
@@ -284,8 +328,8 @@ def test_http11_early_hints():
 
     https://datatracker.ietf.org/doc/rfc8297/
     """
-    origin = Origin(b"https", b"example.com", 443)
-    stream = MockStream(
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
         [
             b"HTTP/1.1 103 Early Hints\r\n",
             b"Link: </style.css>; rel=preload; as=style\r\n",
@@ -300,7 +344,7 @@ def test_http11_early_hints():
             b"<html>Hello, world! ...</html>",
         ]
     )
-    with HTTP11Connection(
+    with httpcore.HTTP11Connection(
         origin=origin, stream=stream, keepalive_expiry=5.0
     ) as conn:
         response = conn.request(
@@ -310,3 +354,27 @@ def test_http11_early_hints():
         )
         assert response.status == 200
         assert response.content == b"<html>Hello, world! ...</html>"
+
+
+
+def test_http11_header_sub_100kb():
+    """
+    A connection should be able to handle a http header size up to 100kB.
+    """
+    origin = httpcore.Origin(b"https", b"example.com", 443)
+    stream = httpcore.MockStream(
+        [
+            b"HTTP/1.1 200 OK\r\n",  # 17
+            b"Content-Type: plain/text\r\n",  # 43
+            b"Cookie: " + b"x" * (100 * 1024 - 72) + b"\r\n",  # 102381
+            b"Content-Length: 0\r\n",  # 102400
+            b"\r\n",
+            b"",
+        ]
+    )
+    with httpcore.HTTP11Connection(
+        origin=origin, stream=stream, keepalive_expiry=5.0
+    ) as conn:
+        response = conn.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b""
