@@ -8,9 +8,10 @@ but in a plain old multi-threaded context.
 We don't do any smarts around cancellations, or managing exceptions from
 childen, because we don't need that for our use-case.
 """
+
 import threading
 from types import TracebackType
-from typing import List, Type
+from typing import Any, Callable, List, Optional, Type
 
 
 class Nursery:
@@ -22,16 +23,16 @@ class Nursery:
 
     def __exit__(
         self,
-        exc_type: Type[BaseException] = None,
-        exc_value: BaseException = None,
-        traceback: TracebackType = None,
+        exc_type: Optional[Type[BaseException]] = None,
+        exc_value: Optional[BaseException] = None,
+        traceback: Optional[TracebackType] = None,
     ) -> None:
         for thread in self._threads:
             thread.start()
         for thread in self._threads:
             thread.join()
 
-    def start_soon(self, func, *args):
+    def start_soon(self, func: Callable[..., object], *args: Any) -> None:
         thread = threading.Thread(target=func, args=args)
         self._threads.append(thread)
 
